from __future__ import annotations
import time
import logging
import numpy as np

from typing import Callable, Optional, Tuple
from abc import ABC, abstractmethod
from pathlib import Path
from cv_pom.cv_pom import POM, POMElement

empty_pom_element = POMElement("", "", (0, 0), (0, 0), (0, 0), (0, 0, 0, 0), 0, {})

# create logger
logger = logging.getLogger('cv_pom logger')
logger.setLevel(logging.DEBUG)
# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
# create formatter
formatter = logging.Formatter('[%(asctime)s] - %(levelname)s - %(message)s')
# add formatter to ch
ch.setFormatter(formatter)
# add ch to logger
logger.addHandler(ch)


class CVPOMDriverElement(POMElement):
    """This class extends POMElement by adding methods for interactions"""

    def __init__(self, element: POMElement, query: dict, driver: "CVPOMDriver") -> None:
        super().__init__(**element.as_dict())  # Init POMElement part (parent class)
        self._query = query
        self._driver = driver

    def click(self, timeout=10, offset=(0, 0), times=1, interval=0, button="PRIMARY") -> CVPOMDriverElement:
        """Click in the center of an element.

        Will wait for element to be visible first.

        Args:
            timeout: Max wait time in seconds. Defaults to 10.
            interval: interval when click times is more than 1
            button: button to use for clicking
            times: defaults to 1, and 2 performs double click for those frameworks that allows it
            offset: Offset from the coordinates of the element (AX, AY)

        Returns:
            CVPOMDriverElement
        """
        self.wait_visible(timeout)
        x, y = self.center
        ax, ay = offset
        logger.info(
            f"action: click - element coords: {self.center} - element label: \"{self.label}\" - element attrs: {self.attrs}"
        )
        self._driver._click_coordinates(x + ax, y + ay, times, interval, button)

        return self

    def wait_visible(self, timeout=10) -> CVPOMDriverElement:
        """Wait until element is visible

        Args:
            timeout: Max wait time in seconds. Defaults to 10.

        Returns:
            CVPOMDriverElement
        """
        if self.center == (0, 0):
            elements: list[CVPOMDriverElement] = self._driver.wait_until(
                lambda: self._driver.elements(self._query),
                lambda els: len(els),
                f"Element '{self._query}' not found after {timeout}s",
                timeout
            )
            self.__init__(elements[0], self._query, self._driver)

        logger.info(
            f"action: wait_visible - element coords: {self.center} - "
            f"element label: \"{self.label}\" - element attrs: {self.attrs}"
        )

        return self

    def wait_not_visible(self, timeout=10) -> CVPOMDriverElement:
        """Wait until element is not visible

        Args:
            timeout: Max wait time in seconds. Defaults to 10.

        Returns:
            CVPOMDriverElement
        """
        self._driver.wait_until(
            lambda: self._driver.elements(self._query),
            lambda els: len(els) == 0,
            f"Element '{self._query}' still visible after {timeout}s",
            timeout
        )

        logger.info(
            f"action: wait_not_visible - element coords: {self.center} - "
            f"element label: \"{self.label}\" - element attrs: {self.attrs}"
        )

        return self

    def send_keys(self, keys: str, offset=(0, 0)) -> CVPOMDriverElement:
        """Send keys.

        Will wait for element to be visible first.

        Args:
            keys: Key sequence (string) to send
            offset: Offset from the coordinates of the element (AX, AY)

        Returns:
            CVPOMDriverElement
        """
        self.click(offset=offset)  # Focus the input element
        logger.info(
            f"action: send_keys - element coords: {self.center} - "
            f"element label: \"{self.label}\" - element attrs: {self.attrs}"
        )
        self._driver._send_keys(keys)

        return self

    def swipe(self, offset: tuple = None, el: CVPOMDriverElement = None) -> CVPOMDriverElement:
        """Swipes or scrolls using coordinates, direction or an element.

        Args:
            el: if is not None then it will scroll from the el1 to the el
            offset: if is not None then it will scroll from the el1 to the offset marked by (x, y)
        Returns:
            CVPOMDriverElement
        """
        el1 = self.wait_visible()
        if offset is not None:
            x, y = el1.center
            delta_x, delta_y = offset
            x_end, y_end = x + delta_x, y + delta_y
        else:
            x, y = el1.center
            x_end, y_end = el.wait_visible().center
        self._driver._swipe_coordinates(coords=(x, y, x_end, y_end))

        return el1

    def swipe_to(self, direction: str, limit: int = 50) -> CVPOMDriverElement:
        """Swipes or scrolls direction until it finds the element.

        Args:
            direction: if is not none then will scroll/swipe "up", "down", "left" and "right"
            limit: amount of scrolls so that it does not loop infinitely, defaults to 50
        Returns:
            CVPOMDriverElement
        """
        i = 0
        elements = self._driver.elements(self._query)
        while len(elements) == 0 and i < limit:
            self._driver._swipe_coordinates(direction=direction)
            elements = self._driver.elements(self._query)
            i += 1

        el = self.wait_visible()
        return el

    def hover(self, timeout=10, offset=(0, 0)) -> CVPOMDriverElement:
        """Hover in the center of an element.

        Will wait for element to be visible first.

        Args:
            timeout: Max wait time in seconds. Defaults to 10.
            offset: Offset from the coordinates of the element (AX, AY)

        Returns:
            CVPOMDriverElement
        """
        self.wait_visible(timeout)
        x, y = self.center
        ax, ay = offset
        logger.info(
            f"action: hover - element coords: {self.center} - element label: {self.label} - element attrs: {self.attrs}"
        )
        self._driver._hover_coordinates(x + ax, y + ay)

        return self

    def drag_drop(self,
                  end_coords: Optional[Tuple[int, int]] = None,
                  delta: Optional[Tuple[int, int]] = None,
                  duration=0.1
                  ) -> CVPOMDriverElement:
        """Drag and Drop from an element to coordinates or a delta distance in pixels.

        Will wait for element to be visible.

        Args:
            offset: Offset from the coordinates of the element (AX, AY)
            end_coords: coordinates to end (optional)
            delta: distance in pixels from the element (optional)
            duration: duration for the action to take place

        Returns:
            CVPOMDriverElement
        """
        if end_coords is None and delta is None:
            logger.error("action: drag_drop - either end_coords or delta must be specified - action not performed")
            return self

        if end_coords is None:
            self.wait_visible()
            x, y = self.center
            delta_x, delta_y = delta
            x_end, y_end = x + delta_x, y + delta_y
        else:
            self.wait_visible()
            x, y = self.center
            x_end, y_end = end_coords

        logger.info(f"action: drag_drop - start coords: {(x, y)} - end coords: {(x_end, y_end)}")
        self._driver._drag_drop(x, y, x_end, y_end, duration)

        return self

    def drag_drop_to(self,
                     start_coords: Optional[Tuple[int, int]] = None,
                     delta: Optional[Tuple[int, int]] = None,
                     duration=0.1
                     ) -> CVPOMDriverElement:
        """Drag and Drop from coordinates or a delta distance in pixels to an element.

        Will wait for element to be visible.

        Args:
            offset: Offset from the coordinates of the element (AX, AY)
            start_coords: coordinates to start from (optional)
            delta: distance in pixels from the element (optional)
            duration: duration for the action to take place

        Returns:
            CVPOMDriverElement
        """
        if start_coords is None and delta is None:
            logger.error("action: drag_drop - either start_coords or delta must be specified - action not performed")
            return self

        if start_coords is None:
            self.wait_visible()
            x_end, y_end = self.center
            delta_x, delta_y = delta
            x, y = x_end + delta_x, y_end + delta_y

        else:
            self.wait_visible()
            x_end, y_end = self.center
            x, y = start_coords

        logger.info(f"action: drag_drop - start coords: {(x, y)} - end coords: {(x_end, y_end)}")
        self._driver._drag_drop(x, y, x_end, y_end, duration)

        return self

    def press(self,
              coords: Optional[Tuple[int, int]] = None, duration=0.1, timeout=10, offset=(0, 0), button="PRIMARY"
              ) -> CVPOMDriverElement:
        """Drag and Drop from an element to coordinates or a delta distance in pixels.

        Will wait for element to be visible.

        Args:
            offset: Offset from the coordinates of the element (AX, AY)
            coords: coordinates to click (optional)
            offset: distance in pixels from the element (optional)
            duration: duration for the action to take place
            timeout: Max wait time in seconds. Defaults to 10.
            button: button to use for clicking
        Returns:
            CVPOMDriverElement
        """

        if coords is None:
            self.wait_visible(timeout=timeout)
            x, y = self.center
        else:
            x, y = coords

        ax, ay = offset
        x, y = x + ax, y + ay
        logger.info(f"action: press - coords: {(x, y)} - duration - {duration}")
        self._driver._drag_drop(x, y, x, y, duration, button)

        return self


class CVPOMPageDriver:
    def __init__(self, pom: POM, driver: "CVPOMDriver", **kwargs) -> None:
        self._pom = pom
        self._driver = driver
        self.kwargs = kwargs

    def element(self, query: dict) -> CVPOMDriverElement:
        elements = self._pom.get_elements(query)

        if len(elements):
            return CVPOMDriverElement(elements[0], query, self._driver)

        # Even if element wasn't found, returns an Element. It can still be used to wait.
        return CVPOMDriverElement(empty_pom_element, query, self._driver)

    def elements(self, query: dict) -> list[CVPOMDriverElement]:
        elements = self._pom.get_elements(query)

        return [CVPOMDriverElement(el, query, self._driver) for el in elements]


class CVPOMDriver(ABC):
    """Driver class used to find elements in the POM"""

    def __init__(self, model_path: Path | str, **kwargs) -> None:
        self._pom = POM(model_path)
        self.kwargs = kwargs

    def element(self, query: dict) -> CVPOMDriverElement:
        """Get a single element.

        This doesn't wait for element to be visible and will not fail.
        The returned CVPOMDriverElement object will be interactable regardless
        of the existence of the element at the time of calling this method.

        Args:
            query: Query dictionary

        Returns:
            CVPOMDriverElement
        """
        ocr = None
        if "text" in query:
            ocr = {'paragraph': False}
            if self.kwargs and self.kwargs['ocr']:
                ocr = self.kwargs['ocr']
        pom = self._pom.convert_to_cvpom(self._get_screenshot(), ocr)

        return CVPOMPageDriver(pom, self).element(query)

    def elements(self, query: dict) -> list[CVPOMDriverElement]:
        """Get a list of elements.

        Args:
            query: Query dictionary

        Returns:
            list[CVPOMDriverElement]: List of elements
        """
        ocr = None
        if "text" in query or "ocr_element" in query:
            ocr = {'paragraph': False}
            if self.kwargs and self.kwargs['ocr']:
                ocr = self.kwargs['ocr']
        pom = self._pom.convert_to_cvpom(self._get_screenshot(), ocr)

        return CVPOMPageDriver(pom, self).elements(query)

    def get_page(self) -> CVPOMPageDriver:
        """Get full page POM.

        Args:

        Returns:
            POM: full CV_POM
        """
        if self.kwargs:
            pom = self._pom.convert_to_cvpom(self._get_screenshot(), self.kwargs['ocr'])
        else:
            pom = self._pom.convert_to_cvpom(self._get_screenshot())

        return CVPOMPageDriver(pom, self)

    def wait_until(self, func: Callable, condition: Callable, error_msg: str, timeout: float):
        """Wait until conditions passes and function is not raising errors.

        Args:
            func: Function to execute
            condition: Condition that needs to be satisfied (uses return value of "func" as parameter)
            error_msg: Error message in case of failure
            timeout: Timeout in seconds

        Raises:
            Exception: Timeout

        Returns:
            Any: Whatever the given function returns
        """
        error = None
        timeout_start = time.time()
        while time.time() < timeout_start + timeout:
            try:
                error = None
                result = func()
                if condition(result):
                    return result
            except Exception as e:
                error = e

        if error:
            raise Exception(f"{error_msg}, error: {error}")
        else:
            raise Exception(f"{error_msg}")

    @abstractmethod
    def _get_screenshot(self) -> np.ndarray:
        pass

    @abstractmethod
    def _click_coordinates(self, x: int, y: int, times=1, interval=0, button="PRIMARY"):
        pass

    @abstractmethod
    def _send_keys(self, keys: str):
        pass

    @abstractmethod
    def _swipe_coordinates(self, coords: tuple = None, direction: str = None, duration: float = 0.1):
        pass

    @abstractmethod
    def _hover_coordinates(self, x: int, y: int):
        pass

    @abstractmethod
    def _drag_drop(self, x: int, y: int, x_end: int, y_end: int, duration=0.1, button="PRIMARY"):
        pass



import cv2 as cv
import numpy as np
import math
import os
import re
import serial
import pygame
import time
import config.device_params as device_params_module
# Mobile device dimensions example:
#   device_params = {
#        "default": {
#            'width': 1080,
#            'height': 2640,
#           'physical_width': 38.0,
#           'physical_height': 102.0,
#           'center_x': 0.0,
#           'center_y': 0.0
#        }
#   }

from cv_pom.cv_pom_driver import CVPOMDriver
from importlib import reload
from pathlib import Path
from config.printer_config import dimensions, SERIAL_PORT, BAUD_RATE 
# SERIAL_PORT is the connection of your device to the printer, and can be found with 'ls /dev/tty.*' terminal command
# BAUD_RATE is set to 115200
# dimensions have the values of the printer bed size in mm
#   Example:
#   dimensions = {
#       "default":{
#           "x": 460, # In mm
#           "y": 460 
#       },
#   }


class PrinterCVPOMDriver(CVPOMDriver):
    def __init__(self, model_path: str | Path, appium_driver, device: str, printer: str, orientation: str, **kwargs) -> None:
        super().__init__(model_path, **kwargs)
        self._driver = appium_driver 
        
        self._ser = None
        self._ser = self.get_connection()
        
        self._device, self._printer = self._load_configs(device, printer)
        self._orientation = orientation
        
        # Default position is the center of the base plate
        self._printer_center_x = self._printer["x"]/2
        self._printer_center_y = self._printer["y"]/2 
        
        if(self._device["center_x"] != 0.0):
            self._printer_center_x = self._device["center_x"]
            self._printer_center_y = self._device["center_y"] 
     
    def _load_configs(self, device: str, printer: str):
         # Get device configuration
        if device not in device_params_module.device_params:
            raise KeyError(f"Device '{device}' not found in device_params")
        device_config = device_params_module.device_params[device]

        # Get printer configuration
        if printer not in dimensions:
            raise KeyError(f"Printer '{printer}' not found in dimensions")
        printer_config = dimensions[printer]

        return device_config, printer_config
    
    def _screen_to_printer(self, x_screen: float, y_screen: float) -> tuple[int, int]:
        screen_width = self._device["width"]
        screen_height = self._device["height"]
        device_width = self._device["physical_width"]
        device_height = self._device["physical_height"]
        
        
        printer_x_center = self._printer_center_x
        printer_y_center = self._printer_center_y 

        # Adjust for orientation
        if self._orientation == "landscape":
            # Swap width/height and rotate coordinates
            screen_width, screen_height = screen_height, screen_width
            device_width, device_height = device_height, device_width

            # Convert portrait to landscape
            temp_x = x_screen
            x_screen = y_screen
            y_screen = screen_width - temp_x
        
        # Scale app position (in px) to real device physical position (mm)
        x_mm = (x_screen / screen_width) * device_width
        y_mm = (1 - (y_screen / screen_height)) * device_height  # Invert Y axis with (1 - ..)

        # Shift so center of the device = printer center
        x_mm_offset = x_mm - (device_width / 2)
        y_mm_offset = y_mm - (device_height / 2)

        # Translate device position relative to printer center
        printer_x = printer_x_center + x_mm_offset
        printer_y = printer_y_center + y_mm_offset

        return printer_x, printer_y

    def _get_screenshot(self) -> np.ndarray:
        png = self._driver.get_screenshot_as_png() # Get screenshot via appium

        # Convert PNG bytes to a numpy array
        arr = np.frombuffer(png, dtype=np.uint8)
        img = cv.imdecode(arr, cv.IMREAD_COLOR)
        
        # Save screenshot in screenshots folder
        save_path = os.path.join(os.path.dirname(__file__), "..", "screenshots")
        os.makedirs(save_path, exist_ok=True)
        cv.imwrite(os.path.join(save_path, "latest.png"), img)
        
        return img

    def _hover_coordinates(self, x: int, y: int, wait_idle: bool = True):
        dx, dy = self._screen_to_printer(x, y)
        self.send_gcode(f"G1 X{dx:.2f} Y{dy:.2f} F3000") # Move to position
        if wait_idle:
            self.wait_until_idle()
    
    def _click_coordinates(self, x: int, y: int, times=1, interval=0, button="PRIMARY", wait_idle: bool = True):
        dx, dy = self._screen_to_printer(x, y)
        self.send_gcode(f"G1 X{dx:.2f} Y{dy:.2f}") # Move to position
        self.send_gcode("G1 Z0 F6000") # Go down
        self.send_gcode("G1 Z10 F6000") # Go back up
        print("Tap!")
        if wait_idle:
            self.wait_until_idle()

    def _swipe_coordinates(self, coords: tuple = None, direction: str = None, duration: float = 0.1, wait_idle: bool = True):
        distance = 30 # How far to swipe
        self.send_gcode(f"G1 X{self._printer_center_x} Y{self._printer_center_y}") # Move to center position
        self.send_gcode("G1 Z2 F6000") # Go down
        
        # Swipe in the correct direction
        if direction == "up":
            swipe = self._printer["y"] / 2 + distance
            self.send_gcode(f"G1 Y{swipe} F6000")
        elif direction == "down":
            swipe = self._printer["y"] / 2 - distance
            self.send_gcode(f"G1 Y{swipe} F6000")
        elif direction == "left":
            swipe = self._printer["x"] / 2 - distance
            self.send_gcode(f"G1 X{swipe} F6000")
        elif direction == "right":
            swipe = self._printer["x"] / 2 + distance
            self.send_gcode(f"G1 X{swipe} F6000")
        else:
            print(f"Invalid direction: {direction}")
            return
            
        self.send_gcode("G1 Z10 F6000") # Go back up
        print(f"Swipe {direction}!")
        
        if wait_idle:
            self.wait_until_idle()

    def _drag_drop(self, x: int, y: int, x_end: int, y_end: int, duration=0.1, button="PRIMARY", wait_idle: bool = True):
        dx, dy = self._screen_to_printer(x, y)
        ex, ey = self._screen_to_printer(x_end, y_end)
        self.send_gcode(f"G1 X{dx:.2f} Y{dy:.2f}") # Move to starting position
        self.send_gcode("G1 Z2") # Go down
        self.send_gcode(f"G1 X{ex:.2f} Y{ey:.2f}") # Move to the target position
        self.send_gcode("G1 Z10") # Go back up
        print("Drag and Drop!")
        if wait_idle:
            self.wait_until_idle()
    
    # Not implemented
    def _send_keys(self, keys: str):
        raise NotImplementedError("Keyboard input functionality is not implemented.")
    
    
    ### ------------------------------------------- Public methods -------------------------------------------
    
    # Get the current X,Y,Z position of the printer
    def get_position(self):
        ser = self.get_connection()
        self.send_gcode("M114")
        while True:
            line = ser.readline().decode(errors="ignore").strip()
            if line and "Count" in line:
                # M114 Output: X:230.00 Y:230.00 Z:12.60 E:0.00 Count X:18400 Y:18400 Z:5040
                match = re.search(r"X:([\d\.\-]+)\s+Y:([\d\.\-]+)\s+Z:([\d\.\-]+)", line)
                if match:
                    x = float(match.group(1))
                    y = float(match.group(2))
                    z = float(match.group(3))
                    return x, y, z
    
    # Use to send any gcode command to the printer     
    def send_gcode(self, cmd: str):
        ser = self.get_connection()

        # If cmd is a list or tuple, iterate through it
        if isinstance(cmd, (list, tuple)):
            for line in cmd:
                ser.write((line + "\r\n").encode())
                ser.flush()
                print(f">> {line}")
                time.sleep(0.05)
        
        # Treat it as a single string
        else:
            ser.write((cmd + "\r\n").encode())
            ser.flush()
            print(f">> {cmd}")
            time.sleep(0.05)
    
    # If manual calibration of the device position was done, call this method to update the center position of the device 
    def update_custom_position(self):
        try:
            reload(device_params_module) # Reload the file where the custom positions were added
            if "custom" in device_params_module.device_params:
                self._device = device_params_module.device_params["custom"]
                self._printer_center_x = self._device.get("center_x", self._printer_center_x)
                self._printer_center_y = self._device.get("center_y", self._printer_center_y)
                print(f"Custom calibration loaded: center=({self._printer_center_x:.2f}, {self._printer_center_y:.2f})")
            else:
                print("No 'custom' entry found in device_params.py.")
        except Exception as e:
            print(f"Failed to load custom calibration: {e}")
    
    # Connect to the printer via serial
    def get_connection(self):
        if self._ser is None:
            self._ser = serial.Serial(port= SERIAL_PORT, baudrate= BAUD_RATE)
            time.sleep(2)  # Wait for printer to be ready
            print("Serial connection established.")
        return self._ser
    
    # Close the existing connection
    def close_connection(self):
        if self._ser is not None:
            self._ser.close()
        print("Serial connection closed.")
        
    def update_appium_driver(self, appium_driver):
        self._driver = appium_driver
        
    # Call to make the program wait until the printer has completed all commands
    def wait_until_idle(self, timeout_s: float = 60.0) -> bool:

        ser = self._ser

        # Remove any old data serial data
        old_timeout = ser.timeout
        ser.timeout = 0.1
        while True:
            data = ser.readline()
            if not data:
                break

        # Wait until all moves are done
        self.send_gcode("M400")

        deadline = time.time() + timeout_s
        ser.timeout = 0.5

        while time.time() < deadline:
            line = ser.readline().decode(errors="ignore").strip().lower()
            if not line:
                continue
            if line.startswith("busy"):
                continue
            if line == "ok":
                # Final OK after M400 means motion queue is empty
                ser.timeout = old_timeout
                # Returns a bool if more complex logic is needed during implementation
                return True

        # Timeout
        ser.timeout = old_timeout
        
        # Returns a bool if more complex logic is needed during implementation
        return False
        
        
  # ------------------------------------------- MANUAL CALIBRATION -------------------------------------------
    
    
    def manual_calibration(self):
        self._init_pygame()
        self.send_gcode("G1 Z10")  # lift head a bit

        # Reset saved positions at the start of calibration
        open("custom_position.txt", "w").close()
        clock = pygame.time.Clock()
        last_position = None

        corner_labels = ["Top Left", "Top Right", "Bottom Right", "Bottom Left"]
        coords = {} 
        idx = 0
        print(f"Select {corner_labels[idx]} corner, then left-click.")

        running = True
        while running:
            clock.tick(60)

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False

                elif event.type == pygame.MOUSEMOTION:
                    x, y = event.pos
                    if (last_position is None) or (x, y) != last_position:
                        last_position = (x, y)
                        window_height = self._printer_center_y * 2
                        x, y = event.pos
                        y = window_height - y  # Inverts Y-axis when translating pygame window to printer. Remove this to undo the inversion
                        self._mouse_to_gcode(x, y)

                elif event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
                    # tap down/up
                    self.send_gcode("G1 Z5 F3000")
                    self.send_gcode("G1 Z10 F3000")

                    x, y, _ = self.get_position()
                    print(f"[{corner_labels[idx]}] Current position: X={x:.2f}, Y={y:.2f}")

                    choice = input(f"Save this coordinate for {corner_labels[idx]}? (y/n): ").strip().lower()
                    if choice in ("y", "yes"):
                        label_key = corner_labels[idx].lower()
                        coords[label_key] = (x, y)
                        print("Coordinate saved.")

                        idx += 1
                        if idx >= 4:
                            print("All four corners saved. Exiting manual calibration.")
                            
                            # Ask for device name, default to "custom"
                            device_name = input('Device name to save (default "custom"): ').strip()
                            if not device_name:
                                device_name = "custom"

                            print(f"Saving calibration as '{device_name}'...")
                            self._add_custom_device_to_params(coords=coords, device_name=device_name)
                            print("Exiting manual calibration.")
                            
                            running = False
                        else:
                            print(f"Select {corner_labels[idx]} corner, then left-click.")
                    else:
                        print("Coordinate not saved. Try again for the same corner.")
    
    def _init_pygame(self, title: str = "Manual Control"):
        width = self._printer_center_x*2
        height = self._printer_center_y*2
        pygame.init()
        screen = pygame.display.set_mode((width, height))
        pygame.display.set_caption(title)
        clock = pygame.time.Clock()
        return screen, clock

    def _get_mouse_position(self):
        return pygame.mouse.get_pos()


    def _mouse_to_gcode(self, x: int, y: int):
        
        # Creating new attributes to avoid sending every movement to the printer, only the last
        if not hasattr(self, "_last_move_time"):
            self._last_move_time = 0.0
            self._last_move_pos = None

        debounce_s = 0.15  # How many seconds to wait between sending last command to printer
        now = time.monotonic()

        # If not enough time passed since last method call, skip this one 
        if now - self._last_move_time < debounce_s:
            self._last_move_pos = (x, y)
            return

        # If we had a pending position from previous rapid moves, use that instead
        if self._last_move_pos is not None:
            x, y = self._last_move_pos
            self._last_move_pos = None

        self._last_move_time = now

        gcode = f"G1 X{x} Y{y} F10000"
        self.send_gcode(gcode)

    def _add_custom_device_to_params(self, coords: dict, device_name: str = "custom", params_path="config/device_params.py"):
        
        required = ["top left", "top right", "bottom right", "bottom left"]
        if any(label not in coords for label in required):
            raise ValueError(
                "Expected coordinates for Top Left, Top Right, Bottom Right, Bottom Left"
            )

        tl, tr, br, bl = (
            coords["top left"],
            coords["top right"],
            coords["bottom right"],
            coords["bottom left"],
        )

        # Compute dimensions
        width = math.dist(tl, tr)
        height = math.dist(tl, bl)

        # Compute physical center
        cx = (tl[0] + tr[0] + bl[0] + br[0]) / 4.0
        cy = (tl[1] + tr[1] + bl[1] + br[1]) / 4.0

        
        size = self._driver.get_window_size()
        screen_width = size["width"]
        screen_height = size["height"]

        entry = f"""
        '{device_name}': {{
            'width': {screen_width},
            'height': {screen_height},
            'physical_width': {width:.2f},
            'physical_height': {height:.2f},
            'center_x': {cx:.2f},
            'center_y': {cy:.2f}
        }},
        """

        # Update or append to device_params.py
        if not os.path.exists(params_path):
            with open(params_path, "w", encoding="utf-8") as f:
                f.write("device_params = {" + entry + "\n}\n")
            print(f"Created {params_path} and added '{device_name}' entry.")
            return

        with open(params_path, "r+", encoding="utf-8") as f:
            content = f.read()
            
            pattern = rf"'{re.escape(device_name)}':\s*\{{[^}}]+\}},"
            if f"'{device_name}':" in content:
                print(f"Updating existing '{device_name}' entry...")
                content = re.sub(pattern, entry.strip(), content)
            else:
                # Append before closing
                content = re.sub(r"\}\s*$", entry + "\n}", content)
                
            f.seek(0)
            f.write(content)
            f.truncate()

        print(f"Device '{device_name}' has been added/updated with width={width:.2f} mm, height={height:.2f} mm, center=({cx:.2f}, {cy:.2f})")


from pathlib import Path
from cv_pom.cv_pom_driver import CVPOMDriver
import numpy as np
import cv2 as cv
from PIL import Image

try:
    import pyautogui
except ImportError:
    pass


class DesktopCVPOMDriver(CVPOMDriver):
    """CVPOMDriver adapted for PyAutoGUI framework"""

    def __init__(self, model_path: Path | str, **kwargs) -> None:
        """Initialize the driver

        Args:
            model_path: path to the CVPOM model
            driver: path to the TestUIDriver
        """
        super().__init__(model_path, **kwargs)
        self.resize = 1
        if kwargs and "resize" in kwargs:
            self.resize = kwargs["resize"]

    def _get_screenshot(self) -> np.ndarray:
        screenshot = pyautogui.screenshot()
        width, height = screenshot.size
        screenshot = screenshot.resize((round(width*self.resize), round(height*self.resize)), Image.LANCZOS)
        pimg = np.array(screenshot)
        return cv.cvtColor(np.array(pimg), cv.COLOR_RGB2BGR)

    def _click_coordinates(self, x: int, y: int, times=1, interval=0, button="PRIMARY"):
        pyautogui.click(x=x, y=y, clicks=times, interval=interval, button=button)

    def _send_keys(self, keys: str):
        pyautogui.write(keys)

    def _swipe_coordinates(self, coords: tuple = None, direction: str = None, duration=None):
        # duration var doesn't have an effect in this driver

        if coords:
            print("Coordinates scroll not supported")
        if direction == 'up':
            pyautogui.scroll(1)
        elif direction == 'down':
            pyautogui.scroll(-1)
        elif direction == 'left':
            pyautogui.hscroll(1)
        elif direction == 'right':
            pyautogui.hscroll(-1)

    def _hover_coordinates(self, x: int, y: int):
        pyautogui.moveTo(x, y)

    def _drag_drop(self, x: int, y: int, x_end: int, y_end: int, duration=0.1, button="PRIMARY"):
        pyautogui.mouseDown(x, y, button=button)
        pyautogui.moveTo(x, y, duration=duration)
        pyautogui.mouseUp(x_end, y_end)



import base64
import inspect
import time
from io import BytesIO
from pathlib import Path
from numpy import ndarray
import cv2 as cv
from cv_pom.cv_pom_driver import CVPOMDriver
from PIL import Image
import numpy as np

try:
    from testui.support.logger import log_info
    from testui.support.testui_driver import TestUIDriver
    from selenium.webdriver.common.actions.action_builder import ActionBuilder
    from selenium.webdriver.common.actions import interaction
    from selenium.webdriver.common.actions.pointer_input import PointerInput
except ImportError:
    pass


class TestUICVPOMDriver(CVPOMDriver):
    """CVPOMDriver adapted for Py-TestUI framework"""

    def __init__(self, model_path: Path | str, driver: TestUIDriver, **kwargs) -> None:
        """Initialize the driver

        Args:
            model_path: path to the CVPOM model
            driver: path to the TestUIDriver
        """
        super().__init__(model_path, **kwargs)
        self._driver = driver
        self.resize = 1
        if kwargs and "resize" in kwargs:
            self.resize = kwargs["resize"]

    def _get_screenshot(self) -> ndarray:
        driver = self._driver.get_driver  # Deprecated property 1.2.1 python-testui
        if inspect.ismethod(self._driver.get_driver):
            driver = self._driver.get_driver()
        image = driver.get_screenshot_as_base64()
        sbuf = BytesIO()
        sbuf.write(base64.b64decode(str(image)))
        pimg = Image.open(sbuf)
        size = self._driver.get_driver().get_window_size()
        if self._driver.device_udid is not None:
            h, w = size["width"], size["height"]
        else:
            h, w = pimg.size
        if self.resize != 1:
            width, height = pimg.size
            h, w = width*self.resize, height*self.resize
        pimg = pimg.resize((h, w), Image.LANCZOS)
        return cv.cvtColor(np.array(pimg), cv.COLOR_RGB2BGR)

    def _click_coordinates(self, x: int, y: int, times=1, interval=0, button="PRIMARY"):
        driver = self._driver.get_driver  # Deprecated property 1.2.1 python-testui
        if inspect.ismethod(self._driver.get_driver):
            driver = self._driver.get_driver()
        actions = ActionBuilder(
            driver,
            mouse=PointerInput(interaction.POINTER_TOUCH, "touch"),
        )

        for i in range(times):
            actions.pointer_action.move_to_location(x=x, y=y)
            actions.pointer_action.click()
            actions.perform()
            time.sleep(interval)

    def _send_keys(self, keys: str):
        self._driver.actions().send_keys(keys).perform()

    def _swipe_coordinates(self, coords: tuple = None, direction: str = None, duration: float = 0.2):
        actions = self._driver.actions()

        if coords is not None:
            x, y, x_end, y_end = coords
        else:
            h, w, ch = self._get_screenshot().shape
            if direction.lower() == "down":
                x, y, x_end, y_end = int(w / 2), int(4 * h / 6), int(w / 2), int(2 * h / 6)
            elif direction.lower() == "up":
                x, y, x_end, y_end = int(w / 2), int(2 * h / 6), int(w / 2), int(4 * h / 6)
            elif direction.lower() == "left":
                x, y, x_end, y_end = int(w / 4), int(h / 2), int(3 * w / 4), int(h / 2)
            elif direction.lower() == "right":
                x, y, x_end, y_end = int(3 * w / 4), int(h / 2), int(w / 4), int(h / 2)
            else:
                raise Exception(f"direction has to be one of this: down, up, left, right. Was {direction}")

        if self._driver.device_udid is None:
            delta_x = x - x_end
            delta_y = y - y_end
            log_info(f"swiping deltas: {delta_x}, {delta_y}")
            actions.w3c_actions.wheel_action.scroll(delta_x=delta_x, delta_y=delta_y)
            actions.perform()
            return

        log_info(f"swiping coordinates: {x}, {y}, {x_end}, {y_end}")
        driver = self._driver.get_driver  # Deprecated property 1.2.1 python-testui
        if inspect.ismethod(self._driver.get_driver):
            driver = self._driver.get_driver()

        if self._driver.device_udid is not None:
            driver.update_settings({"appium:settings[animationCoolOffTimeout]": duration})

        actions = ActionBuilder(
            driver,
            mouse=PointerInput(interaction.POINTER_TOUCH, "touch"),
            duration=int(duration * 1000)
        )
        actions.pointer_action.move_to_location(x=x, y=y)
        actions.pointer_action.pointer_down()
        actions.pointer_action.move_to_location(x=x_end, y=y_end)
        actions.pointer_action.pointer_up()
        actions.perform()

    def _hover_coordinates(self, x: int, y: int):
        driver = self._driver.get_driver  # Deprecated property 1.2.1 python-testui
        if inspect.ismethod(self._driver.get_driver):
            driver = self._driver.get_driver()
        actions = ActionBuilder(
            driver,
            mouse=PointerInput(interaction.POINTER_TOUCH, "touch"),
        )

        actions.pointer_action.move_to_location(x=x, y=y)
        actions.perform()

    def _drag_drop(self, x: int, y: int, x_end: int, y_end: int, duration=0.1, button="PRIMARY"):
        driver = self._driver.get_driver  # Deprecated property 1.2.1 python-testui
        if inspect.ismethod(self._driver.get_driver):
            driver = self._driver.get_driver()
        if self._driver.device_udid is not None:
            driver.update_settings({"appium:settings[animationCoolOffTimeout]": duration})
        actions = ActionBuilder(
            driver,
            mouse=PointerInput(interaction.POINTER_TOUCH, "touch"),
            duration=int(duration * 1000)
        )

        actions.pointer_action.move_to_location(x=x, y=y)
        actions.pointer_action.pointer_down(button=button)
        actions.pointer_action.pause(duration)
        actions.pointer_action.move_to_location(x=x_end, y=y_end)
        actions.pointer_action.pointer_up(button=button)
        actions.perform()



import time

from pathlib import Path
from cv_pom.cv_pom_driver import CVPOMDriver
import subprocess as sp
import cv2 as cv

from numpy import ndarray


class AdbCVPOMDriver(CVPOMDriver):
    def __init__(self, model_path: Path | str, url: str = None, udid: str = None) -> None:
        super().__init__(model_path)
        self.__udid = ""
        if udid is not None:
            self.__udid = f"-s {udid} "
        if url is not None:
            sp.run(f"adb {self.__udid}shell am start -a android.intent.action.VIEW -d {url}", shell=True)

    def _click_coordinates(self, x: int, y: int, times=1, interval=0, button="PRIMARY"):
        for _ in range(times):
            sp.run(f"adb {self.__udid}shell input tap {x} {y}", shell=True)
            time.sleep(interval)

    def _get_screenshot(self) -> ndarray:
        sp.run(f"adb {self.__udid}exec-out screencap -p > screen.png", shell=True)
        return cv.imread("screen.png")

    def _send_keys(self, keys: str):
        sp.run(f"adb {self.__udid}shell input text \"{keys}\"", shell=True)

    def _drag_drop(self, x: int, y: int, x_end: int, y_end: int, duration=0.1, button="PRIMARY"):
        sp.run(f"adb {self.__udid}shell input touchscreen swipe {x} {y} {x_end} {y_end}", shell=True)

    def _hover_coordinates(self, x: int, y: int):
        print("doesn't exist for ADB driver")

    def _swipe_coordinates(self, coords: tuple = None, direction: str = None, duration=None):
        # duration var doesn't have an effect in this driver

        if coords is not None:
            x, y, x_end, y_end = coords
        else:
            h, w, ch = self._get_screenshot().shape
            if direction.lower() == "down":
                x, y, x_end, y_end = int(w / 2), int(4 * h / 6), int(w / 2), int(2 * h / 6)
            elif direction.lower() == "up":
                x, y, x_end, y_end = int(w / 2), int(2 * h / 6), int(w / 2), int(4 * h / 6)
            elif direction.lower() == "left":
                x, y, x_end, y_end = int(w / 4), int(h / 2), int(3 * w / 4), int(h / 2)
            elif direction.lower() == "right":
                x, y, x_end, y_end = int(3 * w / 4), int(h / 2), int(w / 4), int(h / 2)
            else:
                raise Exception(f"direction has to be one of this: down, up, left, right. Was {direction}")

        sp.run(f"adb {self.__udid}shell input touchscreen swipe {x} {y} {x_end} {y_end}", shell=True)



from __future__ import annotations
import dataclasses
from dataclasses import dataclass
import json
from pathlib import Path
from typing import Any
import numpy as np
from ultralytics import YOLO
import cv2 as cv
import easyocr
from ultralytics.engine.results import Results


@dataclass
class QueryValue:
    value: str
    case_sensitive: bool = True
    contains: bool = False


@dataclass
class Query:
    label: QueryValue | None = None
    text: QueryValue | None = None
    parent: Query | None = None
    child: Query | None = None
    left: Query | None = None
    right: Query | None = None
    up: Query | None = None
    down: Query | None = None


@dataclass
class POMElement:
    id: str
    label: str
    coords_tl: tuple[int, int]
    coords_br: tuple[int, int]
    center: tuple[int, int]
    bounding_rect: tuple[int, int, int, int]
    confidence: float
    attrs: dict[str, Any]

    def as_dict(self):
        return dataclasses.asdict(self)

    def __eq__(self, other):
        if not isinstance(other, POMElement):
            return False
        return (self.id == other.id and self.label == other.label and
                self.coords_tl == other.coords_tl and self.coords_br == other.coords_br and
                self.center == other.center and self.bounding_rect == other.bounding_rect and
                self.confidence == other.confidence and self.attrs == other.attrs)

    def __hash__(self):
        return hash((self.id, self.label, self.coords_tl, self.coords_br,
                     self.center, self.bounding_rect, self.confidence,
                     frozenset(self.attrs.items())))


class POM:
    """CV POM (Page Object Model) class.

    Used to convert image source into POM and then query the elements.
    """

    def __init__(self, model_path: str | Path) -> None:
        self.model = YOLO(model_path)
        self.elements: list[POMElement] = []
        self.annotated_frame: np.ndarray = None
        self._reader = easyocr.Reader(['en'])

    def convert_to_cvpom(self, source, ocr_props=None) -> POM:
        """Convert image to CV POM.

        Args:
            source: Image source, supports same data types as YOLO "predict" method
            ocr_props: properties for OCR to read text. Defaults to None.

        Returns:
            POM: self
        """
        self.elements = []
        self.annotated_frame = None

        #  Object detection
        self.elements = self._object_detection(source)

        # Optical character recognition
        if ocr_props:
            ocr_props_comb = {'image': source, 'batch_size': 3, 'canvas_size': 1200, 'paragraph': False}
            for prop in ocr_props:
                ocr_props_comb[prop] = ocr_props[prop]
            ref_elements_txt = []  # [([Coordinates], Text)]
            output_txts = self._reader.readtext(**ocr_props_comb)
            for output_txt in output_txts:
                coordinates = [output_txt[0][0][0], output_txt[0][0][1], output_txt[0][2][0], output_txt[0][2][1]]
                ref_elements_txt.append([[int(x) for x in coordinates], output_txt[1]])
            # overlapping
            for i in range(len(self.elements)):
                output_overlap = self._find_overlapping_rects([
                    self.elements[i].coords_tl[0],
                    self.elements[i].coords_tl[1],
                    self.elements[i].coords_br[0],
                    self.elements[i].coords_br[1]], ref_elements_txt)
                final_text = ''
                for output_txt in output_overlap:
                    final_text += output_txt[1] + ' '
                self.elements[i].attrs['text'] = final_text[:-1]
            # non - overlapping
            ocr_elements = self._find_non_overlapping_rects(self.elements, ref_elements_txt)
            for ocr_element in ocr_elements:
                id_n = len(self.elements)
                ocr_element.id = str(id_n)
                self.annotated_frame = cv.rectangle(
                    self.annotated_frame,
                    pt1=ocr_element.coords_br,
                    pt2=ocr_element.coords_tl,
                    color=(0, 255, 0),
                    thickness=3,
                )
                x, y = ocr_element.coords_tl
                cv.putText(
                    self.annotated_frame, 'ocr_element', (x, y - 10), cv.FONT_HERSHEY_SIMPLEX, 0.9, (0, 255, 0), 3
                )
                self.elements.append(ocr_element)

        return self

    def _object_detection(self, source) -> list[POMElement]:
        """Retrieves all the elements using Object Detection ML Algorithm.
        Args:
            source: Image
        Returns:
            list[POMElement]: List of elements
        """
        elements: list[POMElement] = []

        results: list[Results] = self.model.predict(source, verbose=False)

        for result in results:
            self.annotated_frame = result.plot()
            for i in range(len(result.boxes.data.tolist())):
                data = result.boxes.data.tolist()[i]
                xmin, ymin, xmax, ymax, confidence, class_id = data
                tl = (int(xmin), int(ymin))
                br = (int(xmax), int(ymax))

                element = POMElement(
                    id=str(i),
                    label=results[0].names[class_id],
                    coords_tl=tl,
                    coords_br=br,
                    center=(
                        (br[0] + tl[0]) // 2,
                        (br[1] + tl[1]) // 2,
                    ),
                    bounding_rect=cv.boundingRect(np.array([tl, br])),
                    confidence=confidence,
                    attrs={},
                )

                elements.append(element)

        return elements

    def get_elements(self, query_dict: dict | None = None) -> list[POMElement]:
        """Get elements using a query.

        Query examples:
        - select by exact label:                {"label": "my-label"}
        - select by label containing substring: {"label": {"value: "my-label", "contains": True}}
        - select by label not case sensitive:   {"label": {"value: "my-label", "case_sensitive": False}}
        - select by exact text:                 {"text": "my-text"}
        - select by exact label and text:       {"label": "my-label", "text": "my-text"}

        When not specified, "case_sensitive" defaults to True and "contains" defaults to False.

        Args:
            query_dict: Query dictionary. Pass None to select all elements. Defaults to None.

        Returns:
            list[POMElement]: List of elements
        """
        if query_dict is None:
            return self.elements

        query = self._parse_query(query_dict)

        return self._filter_query(query)

    def to_json(self, indent=None) -> str:
        return json.dumps([el.as_dict() for el in self.elements], indent=indent)

    def _filter_query(self, query: Query) -> list[POMElement]:
        filtered: list[POMElement] = []

        for element in self.elements:
            label = element.label
            if not self._check_query_value(label, query.label):
                continue

            if query.text is not None and "text" not in element.attrs:
                continue

            if "text" in element.attrs:
                text = element.attrs["text"]
                if not self._check_query_value(text, query.text):
                    continue

            filtered.append(element)

        filtered_in: list[POMElement] = []
        if query.child is not None:
            filtered_child = self._filter_query(query.child)
            for element in filtered:
                for el in self._find_contained_rects(element, filtered_child):
                    filtered_in.append(el)
            filtered = filtered_in

        filtered_in = []
        if query.parent is not None:
            filtered_parent = self._filter_query(query.parent)
            for element in filtered_parent:
                if len(self._find_contained_rects(element, filtered)) > 0:
                    filtered_in.append(element)
            filtered = list(dict.fromkeys(filtered_in))

        filtered_in = []
        if query.left is not None:
            filtered_left = self._filter_query(query.left)
            for element in filtered:
                for el in self._find_rects_sides(element, filtered_left, "left"):
                    filtered_in.append(el)
            filtered = list(dict.fromkeys(filtered_in))

        filtered_in = []
        if query.right is not None:
            filtered_right = self._filter_query(query.right)
            for element in filtered:
                for el in self._find_rects_sides(element, filtered_right, "right"):
                    filtered_in.append(el)
            filtered = list(dict.fromkeys(filtered_in))

        filtered_in = []
        if query.up is not None:
            filtered_up = self._filter_query(query.up)
            for element in filtered:
                for el in self._find_rects_sides(element, filtered_up, "up"):
                    filtered_in.append(el)
            filtered = list(dict.fromkeys(filtered_in))

        filtered_in = []
        if query.down is not None:
            filtered_down = self._filter_query(query.down)
            for element in filtered:
                for el in self._find_rects_sides(element, filtered_down, "down"):
                    filtered_in.append(el)
            filtered = list(dict.fromkeys(filtered_in))

        return filtered

    def _check_query_value(self, value: str, query_val: QueryValue | None) -> bool:
        if query_val is None:
            return True

        expected = query_val.value
        if not query_val.case_sensitive:
            value = value.lower()
            expected = expected.lower()

        if query_val.contains:
            if expected not in value:
                return False
        else:
            if expected != value:
                return False

        return True

    def _parse_query(self, query_dict: dict) -> Query:
        """While the user can pass query as a dictionary,
        internally we want to work with concrete types.
        """
        query = Query()
        for [key, item] in query_dict.items():
            if key == "label" or key == "text":
                if isinstance(item, dict):
                    if "value" not in item:
                        raise Exception("Query object doesn't have 'value' field")
                    query_val = QueryValue(item["value"])
                    if "case_sensitive" in item:
                        query_val.case_sensitive = bool(item["case_sensitive"])
                    if "contains" in item:
                        query_val.contains = bool(item["contains"])
                    setattr(query, key, query_val)
                else:
                    setattr(query, key, QueryValue(item))
            else:
                if isinstance(item, dict):
                    query_in = self._parse_query(item)
                    setattr(query, key, query_in)

        return query

    def _do_rect_share_space(self, rect1, rect2) -> bool:
        """Check if rects share space by comparing their coordinates
        rect = [xmin, ymin, xmax, ymax]

        returns bool
        """
        return not (rect1[2] < rect2[0] or  # rect1 is to the left of rect2
                    rect1[0] > rect2[2] or  # rect1 is to the right of rect2
                    rect1[3] < rect2[1] or  # rect1 is above rect2
                    rect1[1] > rect2[3])

    def _do_rect1_contain_rect2(self, rect1, rect2) -> bool:
        """Check if rect1 contains within rect2
        rect = [xmin, ymin, xmax, ymax]

        returns bool
        """
        return (rect2[2] < rect1[2] and  # rect2 xmax is lower than rect1
                rect2[0] > rect1[0] and  # rect2 xmin is bigger than rect1
                rect2[3] < rect1[3] and  # rect2 ymax is lower than rect1
                rect2[1] > rect1[1])

    def _find_contained_rects(self, element_1: POMElement, elements: list[POMElement]) -> list[POMElement]:
        """Example usage:
        rect1 = [xmin, ymin, xmax, ymax]
        list[POMElement]

        returns list
        """
        rect1 = [element_1.coords_tl[0], element_1.coords_tl[1], element_1.coords_br[0], element_1.coords_br[1]]
        contained_elements: list[POMElement] = []

        for element in elements:
            # Check for overlap in x-axis and y-axis
            rect2 = [element.coords_tl[0], element.coords_tl[1], element.coords_br[0], element.coords_br[1]]
            if self._do_rect1_contain_rect2(rect1, rect2):
                # Rectangles overlap, add to the result
                contained_elements.append(element)

        return contained_elements

    def _find_rects_sides(self, element_1: POMElement, elements: list[POMElement], side: str) -> list[POMElement]:
        """Example usage:
        POMElement - [xmin, ymin, xmax, ymax]
        list[POMElement] - [xmin, ymin, xmax, ymax]
        left/right/up/down

        returns list
        """
        contained_elements: list[POMElement] = []

        for element in elements:
            if side == "right":
                if (element.coords_tl[0] > element_1.coords_br[0]
                        and element.coords_tl[1] < element_1.center[1] < element.coords_br[1]):
                    contained_elements.append(element)
            if side == "left":
                if (element_1.coords_tl[0] > element.coords_br[0]
                        and element.coords_tl[1] < element_1.center[1] < element.coords_br[1]):
                    contained_elements.append(element)
            if side == "up":
                if (element_1.coords_tl[1] < element.coords_br[1]
                        and element.coords_tl[0] < element_1.center[0] < element.coords_br[0]):
                    contained_elements.append(element)
            if side == "down":
                if (element_1.coords_br[1] > element.coords_tl[1]
                        and element.coords_tl[0] < element_1.center[0] < element.coords_br[0]):
                    contained_elements.append(element)

        return contained_elements

    def _find_overlapping_rects(self, rect1: list, sq_w_txt: list) -> list:
        """Example usage:
        rect1 = [4, 1, 6, 3]
        sq_w_txt = [[[7, 4, 9, 2], 'text1'], [[11, 8, 13, 6], [5, 2, 7, 1], 'text2']]
        _find_overlapping_rects(rects1, rects2)

        returns list
        """
        overlapping_rects = []

        for sq_txt in sq_w_txt:
            # Check for overlap in x-axis and y-axis
            if self._do_rect_share_space(rect1, sq_txt[0]):
                # Rectangles overlap, add to the result
                overlapping_rects.append(sq_txt)

        return overlapping_rects

    def _find_non_overlapping_rects(self, pom_els: list[POMElement], ocr_els: list) -> list[POMElement]:
        """It will find all those elements from OCR that does not overlap with the
        Yolo recognised elements

        Example usage:
        pom_els = [POMElement(...), ...]
        ocr_els = [[11, 8, 13, 6], [5, 2, 7, 1], 'text2']]
        _find_non_overlapping_rects(pom_els, ocr_els)

        returns list[POMElement]
        """
        non_overlapping_rects = []

        for ocr_el in ocr_els:
            is_overlapping = False
            for pom_el in pom_els:
                if self._do_rect_share_space(
                        [pom_el.coords_tl[0],
                         pom_el.coords_tl[1],
                         pom_el.coords_br[0],
                         pom_el.coords_br[1]], ocr_el[0]):
                    is_overlapping = True

            if not is_overlapping:
                element = POMElement(
                    id='',
                    label='ocr_element',
                    coords_tl=(ocr_el[0][0], ocr_el[0][1]),
                    coords_br=(ocr_el[0][2], ocr_el[0][3]),
                    center=(
                        (ocr_el[0][0] + ocr_el[0][2]) // 2,
                        (ocr_el[0][1] + ocr_el[0][3]) // 2,
                    ),
                    bounding_rect=(
                        cv.boundingRect(np.array([(ocr_el[0][0], ocr_el[0][1]), (ocr_el[0][2], ocr_el[0][3])]))
                    ),
                    confidence=1.0,
                    attrs={'text': ocr_el[1]},
                )
                non_overlapping_rects.append(element)

        return non_overlapping_rects
